#ifndef IDLC_CPP_H
#define IDLC_CPP_H

#include <fstream>
#include <list>

#include <lang.h>
#include <util.h>

using std::ofstream;
using std::ostream;

// Output the namespace-qualified symbol name, including
// user namespaces.   A prefix for the final component (i.e. the
// actual name) can optionally be specified.

void cpp_output_name(ostream &file, Symbol *sym, const char *prefix = "");

// Output the C++ type, not including the array, including
// a trailing space.

void cpp_output_type(ostream &file, Type *t, bool array, bool is_mutable);
void cpp_output_type(ostream &file, CompiledBasicType &t, bool is_mutable);

// Output the param's fully-qualified type, a space, and the name
// of the param.

void cpp_output_one_param(ostream &file, Param *p, bool is_server,
                          bool is_copy = false);
	
// Each IDL namespace is mapped to a C++ header file,
// represented by a CPPFile.

class CPPFile : public LangCallback {
	Indent indent;
	string dirname;
	ofstream file;
	UserNameSpace *ns;
	StrList *fqname;

	//// Interface caller-side members:
	
	// Outputs the body of the internal _i_<ifacename> struct,
	// containing a vtable pointer (and definition).

	void output_internal(Interface *sym);
	
	// Outputs the user-visible class wrapper that is used to
	// allow implicit upcasts.
	
	void output_wrapper(Interface *sym);

	// Output implementations of the cast methods prototyped in the wrapper.
	// The implementation is delayed to avoid circular dependency problems.

	void output_casts(Interface *sym);

	// Output the downcast and implicit upcast methods for
	// the given interface/superinterface pair.

	static void wrapper_cast_callback(Interface *iface, void *arg);
	void output_downcast(Interface *iface, Interface *super);
	void output_upcast(Interface *iface, Interface *super);

	static void wrapper_cast_proto_callback(Interface *iface, void *arg);
	void output_downcast_proto(Interface *iface, Interface *super);
	void output_upcast_proto(Interface *iface, Interface *super);

	static void wrapper_method_proto_callback(Interface *iface, void *arg);
	static void wrapper_method_callback(Interface *iface, void *arg);

	void output_iface_tbl(Interface *iface);
	static void tbl_callback(Interface *iface, void *arg);
	void output_iface_tbl_entry(Interface *iface, Interface *super);

	static void output_method_ptrs_callback(Interface *iface, void *arg);
	void output_method_ptrs(Interface *iface);
	void output_methods(Interface *iface, Interface *super, bool prototype);
	void output_method_defs(Interface *iface);

	void CPPFile::output_one_method(Interface *iface, Method *m, 
	                                bool prototype, bool retval);
	void CPPFile::output_one_method_ptr(Method *m, Interface *iface);
	
	static void output_iface_ns(CPPFile *file, NameSpace *sym);

	//// Misc members:

	// Output the static const guid[] value.

	void output_guid(const uint64_t *guid);

	// Output a datum in the given struct, along with any necessary
	// padding.  Return the offset of the next datum.
	
	int output_datum(Struct *ns, Datum *d, int offset);

	static void output_vstruct_ns(CPPFile *file, NameSpace *sym);
	void output_vstruct_info(Struct *sym);
	void output_vstruct_main(Struct *sym);
	void output_struct_ctor(Struct *sym, bool extra_vstruct);
	int output_struct_ctor_rec1(Struct *sym, int num);
	int output_struct_ctor_rec2(Struct *sym, int num);

	void output_bf_elem(Datum *d, int pos, int bits, string &prefix);
	void output_bf(BitField *bf, int bits, int typebits, string &prefix);
		
	// Call ns_in on the symbol in order to declare the containing
	// namespace.  Call ns_out when finished.  ns_out also takes care
	// of inserting the extra newline separating declarations.  The
	// "extra" parameter allows extra output passes to have unique
	// #ifndef names, and should have a trailing underscore.
	
	void ns_in(Symbol *sym, const char *extra = "DEF_");
	void ns_out(Symbol *sym);

	// As above, but also outputs non-user namespaces (optionally
	// including "sym" itself).

	void all_ns_in(Symbol *sym, bool include_self = false,
	               const char *extra = "DEF_");
	void all_ns_out(Symbol *sym, bool include_self = false);

	// Output a protective #ifndef/#define block to keep a type
	// from being declared more than once.  Call ifndef_out()
	// for the #endif.
	
	void ifndef_in(Symbol *sym, const char *extra = "DEF_");
	void ifndef_out();
	
	// Return the namespace-qualified symbol name, excluding
	// user namespaces.  This is used when defining a name
	// which (if it is nested) has been previously forward
	// declared.  A prefix for the final component (i.e. the
	// actual name) can optionally be specified.
	
	String &get_definition_name(Symbol *sym, const char *prefix = "");
	
	// Output the _ns declaration of a struct or interface.
	// If a callback is supplied, it is called inside the
	// namespace, before any namespace members have been
	// emitted.
	
	typedef void (*nsdecl_callback)(CPPFile *file, NameSpace *ns);
	void output_nsdecl(NameSpace *ns, nsdecl_callback cb = NULL);

	void output_aliases_and_types(NameSpace *ns);
	
	// When a top-level struct or interface is about to be generated,
	// it is scanned for other types which need to be generated.  If the
	// full definition is needed, it goes on the need_to_declare list;
	// otherwise, it goes on the need_to_forward_declare list.  When
	// processing the lists, if a node has already been marked as
	// traversed, it is skipped.
	
	// Full declaration is needed for superstructs, superinterfaces,
	// inline structs (once implemented) used, and typedefs used.
	// It is also necessary to fully declare any type which contains
	// a type whose forward declaration is needed, as such types
	// cannot (as far as I know) be forward-declared.
	
	std::list<Symbol *> need_to_declare;
	
	// Forward declaration is needed for types used via a pointer only.
	
	std::list<Symbol *> need_to_forward_declare;
	
	void declare_dependencies(Struct *sym);
	void declare_dependencies(Interface *iface, bool need_obj_def = false);
	void declare_type_dependency(Type *t, bool need_obj_def = false);

	// This is the traversal of the first namespace; any symbol
	// whose last traversal is at least this has been written to
	// at least one of the included namespaces (and/or the
	// current namespace).

	int first_traversal;

	// The traversal indices are also passed as "arg1" to determine
	// which pass to generate.  Higher numbers are performed first.
	
	enum {
		trav_full = 0,     // Full, final output
		trav_obj_def = 1,  // Object struct definition
		trav_obj_stub = 2, // Object pointer stub,
	                      // so it can be used in structs.
		trav_nsdecl = 3,   // _ns declaration
		trav_forward = 4,  // Forward declaration of
	                      // structs/interfaces in nsdecl
	};

	bool pass_needed(Symbol *sym, int pass)
	{
		return first_traversal > sym->traversed[pass];
	}

	void output_pass(Symbol *sym, int pass)
	{
		if (pass_needed(sym, pass)) {
			sym->traversed[pass] = traversal;
			sym->output_lang(this, pass);
		}
	}

	bool do_extra_newline;

	void downscope()
	{
		indent.indent_level++;
		do_extra_newline = false;
	}
	
	void upscope()
	{
		indent.indent_level--;
		do_extra_newline = true;
	}
	
	void extra_newline()
	{
		if (do_extra_newline)
			file << "\n";
		
		do_extra_newline = true;
	}

public:
	CPPFile(UserNameSpace *ns, const char *dir);
	virtual ~CPPFile();

	void output(UserNameSpace *sym, int pass = trav_full, void *arg2 = NULL);
	void output(Struct *sym, int pass = trav_full, void *arg2 = NULL);
	void output(Interface *sym, int pass = trav_full, void *arg2 = NULL);
	void output(BitField *sym, int pass = trav_full, void *arg2 = NULL);
	void output(Enum *sym, int pass = trav_full, void *arg2 = NULL);
	void output(BasicType *sym, int pass = trav_full, void *arg2 = NULL);
	void output(Alias *sym, int pass = trav_full, void *arg2 = NULL);
	void output(TypeDef *sym, int pass = trav_full, void *arg2 = NULL);
	void output(Datum *sym, int pass = trav_full, void *arg2 = NULL);
};

class CPPBinding : public Language {
public:
	CPPBinding()
	{
		name = "C++";
		next = first_lang;
		first_lang = this;
	}

	void output_root(UserNameSpace *ns, const char *dir);
	void output_server(UserNameSpace *ns, const char *dir);
};

#endif
