// idlcomp/languages/c++/main.cc -- C++ IDL binding
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

// FIXME: escape C++ reserved words

#include <cerrno>

#include <sys/types.h>
#include <sys/stat.h>

#include <targets.h>
#include "c++.h"

CPPBinding cppbinding;

CPPFile::CPPFile(UserNameSpace *NS, const char *dir) :
dirname(dir)
{
	indent.indent_level = 0;
	indent.align_spaces = 0;
	do_extra_newline = true;
	ns = NS;

	dirname.append(1, '/');
	dirname.append(ns->name);
	string headername(dirname);
	headername.append(".h");

	file.open(headername.c_str());
	if (!file.is_open()) {
		fprintf(stderr, "Could not create output file \"%s\": %s.\n",
		        headername.c_str(), strerror(errno));

		throw UserError();
	}
	
	if (mkdir(dirname.c_str(), 0777) < 0) {
		fprintf(stderr, "Could not create output directory \"%s\": %s.\n",
		        dir, strerror(errno));

		throw UserError();
	}
	
	file <<   "// " << *ns->get_fq_name()->flatten(".")
	     << "\n// This is a C++ language binding generated by idlc.\n"
	     <<   "// Do not modify this file directly.\n\n";

	fqname = ns->get_fq_name();
	String *fqname_flat = ns->get_fq_name()->flatten("_IDLNS_");
	file <<   "#ifndef IDL_HDR_" << *fqname_flat
	     << "\n#define IDL_HDR_" << *fqname_flat << "\n\n";
	
	file << "#include <orb.h>\n";

	first_traversal = ++traversal;
	assert(first_traversal >= 0);
	
	for (NameSpace::const_iterator i = ns->begin(); i != ns->end(); ++i) {
		Symbol *sym = (*i).second;
		UserNameSpace *uns = dynamic_cast<UserNameSpace *>(sym);
		
		if (uns) {
			file << "#include \"" << **ns->name << '/' << **uns->name << ".h\"\n";
			
			// Process namespaces first, to minimize duplicate definitions
			// if this namespace depends on something declared in a sub
			// namespace.
			
			output(uns);
		}
	}

	for (NameSpace::const_iterator i = ns->begin(); i != ns->end(); ++i) {
		Symbol *sym = (*i).second;
		UserNameSpace *uns = dynamic_cast<UserNameSpace *>(sym);
		
		if (!uns)
			output_pass(sym, trav_nsdecl);
	}

	for (NameSpace::const_iterator i = ns->begin(); i != ns->end(); ++i) {
		Symbol *sym = (*i).second;
		UserNameSpace *uns = dynamic_cast<UserNameSpace *>(sym);
		
		if (!uns)
			output_pass(sym, trav_full);
	}
}

CPPFile::~CPPFile()
{
	file << "#endif\n";
	
	assert(indent.indent_level == 0);
}

void CPPFile::ifndef_in(Symbol *sym, const char *extra)
{
	StrList *fqname = sym->get_fq_name();
	String *fqname_flat = fqname->flatten("_IDLNS_");

	assert(indent.indent_level == 0);

	file << "\n#ifndef IDL_DUP_" << extra << *fqname_flat
	     << "\n#define IDL_DUP_" << extra << *fqname_flat << "\n";
}

void CPPFile::ifndef_out()
{
	file << "\n#endif\n";
}

void CPPFile::ns_in(Symbol *sym, const char *extra)
{
	// Only output user namespaces here; if we're defining a nested type,
	// it will have been forward declared already, and thus the
	// non-user-namespaces can be specified directly in the definition.
	//
	// For non-forward-declarables such as metadata, use all_ns_in.
	
	StrList *ns_fqname = sym->find_toplevel_type()->get_ns()->get_fq_name();

	ifndef_in(sym, extra);

	for (StrList::const_iterator i = ns_fqname->begin();
	     i != ns_fqname->end(); ++i) {
		const String *str = *i;
		file << "namespace " << *str << " {\n";
	}
	
	downscope();
}

void CPPFile::ns_out(Symbol *sym)
{
	upscope();
	assert(indent.indent_level == 0);

	for (Symbol *s = sym->find_toplevel_type()->get_ns(); s != toplevel;
	     s = s->get_ns())
		file << "}";

	ifndef_out();
}

void CPPFile::all_ns_in(Symbol *sym, bool include_self, const char *extra)
{
	NameSpace *ns = sym->find_toplevel_type()->get_ns();
	StrList *ns_fqname = ns->get_fq_name();

	ifndef_in(sym, extra);

	for (StrList::const_iterator i = ns_fqname->begin();
	     i != ns_fqname->end(); ++i) {
		const String *str = *i;
		file << "namespace " << *str << " {\n";
	}
	
	stack<SymbolRef> typens;
	
	if (!include_self)
		sym = sym->get_ns();
	
	for (Symbol *s = sym; s != ns; s = s->get_ns())
		typens.push(s);
	
	while (!typens.empty()) {
		Symbol *s = typens.top();
		typens.pop();
		file << "namespace " << **s->name << "_ns {\n";
	}
	
	downscope();
}

void CPPFile::all_ns_out(Symbol *sym, bool include_self)
{
	upscope();
	assert(indent.indent_level == 0);

	if (!include_self)
		sym = sym->get_ns();

	for (Symbol *s = sym; s != toplevel; s = s->get_ns())
		file << "}";

	ifndef_out();
}

String &CPPFile::get_definition_name(Symbol *sym, const char *prefix)
{
 	NameSpace *ns = sym->get_ns();
	UserNameSpace *uns = dynamic_cast<UserNameSpace *>(ns);
	String *str;
	
	if (!uns) {
		str = &get_definition_name(ns);
		str->append("_ns::");
	} else {
		str = new String();
	}
	
	str->append(prefix);
	str->append(**sym->name);
	return *str;
}

void cpp_output_name(ostream &file, Symbol *sym, const char *prefix)
{
	StrList *sl = sym->get_fq_name("_ns");
	sl->pop_back();

	file << "::" << *sl->flatten("::") << "::" << prefix << **sym->name;
}

void CPPFile::output(UserNameSpace *sym, int pass, void *arg2)
{
	assert(indent.indent_level == 0);
	delete new CPPFile(sym, dirname.c_str());
}

static inline int round_up_bits(int bits)
{
	assert(bits >= 1 && bits <= 64);
	
	if (bits > 32)
		return 64;
	if (bits > 16)
		return 32;
	if (bits > 8)
		return 16;

	return 8;
}

// FIXME: Inline arrays...
void cpp_output_type(ostream &file, Type *t, bool array, bool is_mutable)
{
	if (array) {
		file << "::System::RunTime::";
		
		if (is_mutable)
			file << "Mutable";
		
		file << "Array< ";
	}

	cpp_output_name(file, t);

	if (array)
		file << ">";

	file << ' ';
}

// FIXME: Inline arrays...
void cpp_output_type(ostream &file, CompiledBasicType &t, bool is_mutable)
{
	if (is_array(t)) {
		file << "::System::RunTime::";
		
		if (is_mutable)
			file << "Mutable";
		
		file << "Array< ";
	}

	if (t.flags.field.Bool) {
		// Don't rely on C++ to provide any particular representation
		// of bool...
		file << "uint8_t";
	} else if (t.flags.field.Float) {
		if (t.bits == 32)
			file << "float";
		else
			file << "double";
	} else {
		if (t.flags.field.Unsigned)
			file << 'u';

		file << "int" << round_up_bits(t.bits) << "_t";
	}
	
	if (is_array(t))
		file << '>';
	
	file << ' ';
}

// FIXME: implement padding
int CPPFile::output_datum(Struct *ns, Datum *d, int offset)
{
	if (d->type) {
		cpp_output_type(file, d->type, is_array(d->def.basictype),
		                !d->def.flags.field.Immutable);

		Struct *dtype = dynamic_cast<Struct *>(*d->type);
		if (dtype && !dtype->is_inline())
			file << '*';
	} else {
		cpp_output_type(file, d->def.basictype,
		                !d->def.flags.field.Immutable);
	}

	return offset;
}

void CPPFile::declare_type_dependency(Type *t, bool need_obj_def)
{
	Symbol *toplevel_type = t->find_toplevel_type();

	assert(indent.indent_level == 0);
	
	output_pass(toplevel_type, trav_nsdecl);
	
	if (t == toplevel_type)
		output_pass(t, trav_forward);

	if (dynamic_cast<Interface *>(t))
		output_pass(t, need_obj_def ? trav_obj_def : trav_obj_stub);
	else if (!dynamic_cast<Struct *>(t) || need_obj_def)
		output_pass(t, trav_full);
}

void CPPFile::declare_dependencies(Interface *iface, bool need_obj_def)
{
	for (Interface::methods_iterator i = iface->methods_begin();
	     i != iface->methods_end(); ++i)
	{
		Method *m = *i;
		for (Method::entries_iterator j = m->entries_begin();
		     j != m->entries_end(); ++j)
		{
			Param *p = *j;
			if (p->type)
				declare_type_dependency(p->type, need_obj_def);
		}
	}

	for (Interface::supers_iterator i = iface->supers_begin();
	     i != iface->supers_end(); ++i)
	{
		Interface *super = *i;
		declare_type_dependency(super);
	}
}

void CPPFile::declare_dependencies(Struct *str)
{
	for (NameSpace::const_iterator i = str->begin(); i != str->end(); ++i) {
		Symbol *sym = (*i).second;
		Datum *d = dynamic_cast<Datum *>(sym);
		if (d) {
			if (d->type)
				declare_type_dependency(d->type, d->is_inline());
			
			continue;
		}

		Struct *mstr = dynamic_cast<Struct *>(sym);
		if (mstr) {
			declare_dependencies(mstr);
			continue;
		}
		
		Interface *miface = dynamic_cast<Interface *>(sym);
		if (miface)
			declare_dependencies(miface);
	}
}

void CPPFile::output_guid(const uint64_t *guid64)
{
	const unsigned char *guid = reinterpret_cast<const unsigned char *>(guid64);
	char guidhex[7];
	
	file << indent << "static const __attribute__((unused)) union {\n"
	     << indent << "\tunsigned char c[16];\n"
	     << indent << "\tunsigned long l[];\n"
	     << indent << "} _guid = {\n"
	     << indent << "\t{ ";
	
	for (int i = 0; i < 16; i++) {
		if (i == 8)
			file << '\n' << indent << "\t  ";
		
		sprintf(guidhex, "0x%02x, ", *guid++);
		file << guidhex;
	}

	file << "}\n"
	     << indent << "};\n";

	do_extra_newline = true;
}

void CPPFile::output_nsdecl(NameSpace *ns, nsdecl_callback cb)
{
	// If indent level is not zero, this is a nested struct or interface.
	if (indent.indent_level == 0)
		ns_in(ns, "NS_");
	else
		extra_newline();

	file << indent << "namespace " << **ns->name << "_ns {\n";
	downscope();
	
	if (cb)
		cb(this, ns);
	
	for (NameSpace::const_iterator i = ns->begin(); i != ns->end(); ++i) {
		Symbol *sym = (*i).second;

		assert(pass_needed(sym, trav_forward));
		assert(pass_needed(sym, trav_nsdecl));
		
		// If it's a Method or a non-const Datum, this is a no-op.
		output_pass(sym, trav_forward);
		output_pass(sym, trav_nsdecl);
	}
	
	upscope();
	
	file << indent << "}\n";

	if (indent.indent_level == 1)
		ns_out(ns);
}

void CPPFile::output_aliases_and_types(NameSpace *ns)
{
	for (NameSpace::const_iterator i = ns->begin(); i != ns->end(); ++i) {
		Symbol *sym = (*i).second;
		if (dynamic_cast<Alias *>(sym) || dynamic_cast<Type *>(sym))
			output_pass(sym, trav_full);
	}
}		

void CPPFile::output_vstruct_ns(CPPFile *cpp, NameSpace *sym)
{
	Struct *str = dynamic_cast<Struct *>(sym);
	assert(str);
	assert(str->is_virtual());
	
	cpp->output_guid(str->def.guid);
}

void CPPFile::output_vstruct_info(Struct *sym)
{
	all_ns_in(sym, true, "VINFO_");

	file << indent << "static const __attribute__((unused)) "
	                  "unsigned long *const _guids[] = {\n";
	
	stack<StructRef> supers;
	sym->chainlen = 0;
			
	for (Struct *i = sym; i; i = i->get_super()) {
		sym->chainlen++;
		supers.push(i);
	}
	
	for (int i = 0; i < sym->chainlen; i++) {
		Struct *super = supers.top();
		supers.pop();

		file << indent << '\t';
		cpp_output_name(file, super);
		file << "_ns::_guid.l,\n";
	}

	file << indent << "};\n\n"
	     << indent << "static const __attribute__((unused)) "
	                  "::System::RunTime::VStructInfo _info = {\n"
	     << indent << "\t_guids, " << sym->chainlen << '\n'
	     << indent << "};\n";

	
	all_ns_out(sym, true);
}

void CPPFile::output_vstruct_main(Struct *sym)
{
	assert(sym->is_virtual());
	Struct *super = sym->get_super();

	const char *name = sym->name->c_str();

	if (!super) {
		assert(sym == System_VStruct);
		file << indent << "const ::System::RunTime::VStructInfo "
		                  "*const _infoptr;\n\n";
	}
	
	file << indent << name
	     << "(const ::System::RunTime::VStructInfo *realinfo = &"
	     << name << "_ns::_info) :\n"
	     << indent;

	if (!super)
		file << "_infoptr";
	else
		cpp_output_name(file, super);
	
	file << "(realinfo)\n"
	     << indent << "{\n"
	     << indent << "}\n";

	if (super)
		file << '\n'
		     << indent << "static " << name << " *downcast(::System::VStruct *base)\n"
		     << indent << "{\n"
		     << indent << "\tif (!base)\n"
		     << indent << "\t\treturn NULL;\n\n"
		     << indent << "\tconst ::System::RunTime::VStructInfo *info = base->_infoptr;\n\n"
		     << indent << "\tif (info->chainlen < " << sym->chainlen << ")\n"
		     << indent << "\t\treturn NULL;\n\n"
		     << indent << "\tif (::System::RunTime::guids_equal(info->guids["
		               << sym->chainlen - 1 << "], " << name << "_ns::_guid.l))\n"
		     << indent << "\t\treturn static_cast<" << name << " *>(base);\n\n"
		     << indent << "\treturn NULL;\n"
		     << indent << "}\n";

	do_extra_newline = true;
}

// Output an init method that initializes all the elements in the
// struct; this is useful for throwing exceptions.  Due to the
// annoying, apparently unwaiveable-by-the-struct requirement that any
// struct with a ctor be initialized using the ctor (and thus at
// runtime, not statically), this can't be an actual ctor in
// non-virtual structs.
//
// In virtual structs, there's already a ctor, so it wouldn't
// matter.  It'd generally be best to be consistent and use the
// init method for both, but the main intended use for this is
// throwing exceptions (which are virtual), and it'd be
// unfortunate to cripple 99% of the uses with unnecessary
// syntactic cruft.  The init method will remain available
// for vstructs so that things won't break if a non-vstruct
// gets made virtual (or if a user really wants to be consistent
// between both types in their own code).

void CPPFile::output_struct_ctor(Struct *sym, bool extra_vstruct)
{
	const char *method_name = extra_vstruct ? sym->name->c_str() : " _init";

	file << '\n' << indent << **sym->name;
	indent.align_spaces = sym->name->length() + 1;

	if (!extra_vstruct) {
		file << " &_init";
		indent.align_spaces += 7;
	}

	file << '(';
	output_struct_ctor_rec1(sym, 0);
	
	if (extra_vstruct) {
		Struct *super = sym->get_super();

		file << ",\n" << indent
		     << "const ::System::RunTime::VStructInfo *_realinfo = &"
		     << **sym->name << "_ns::_info) :\n";

		indent.align_spaces = 0;
		file << indent;

		if (!super)
			file << "_infoptr";
		else
			cpp_output_name(file, super);

		file << "(_realinfo";
	}
	
	indent.align_spaces = 0;
	file << ")\n" << indent << "{\n";
	
	indent.indent_level++;

	output_struct_ctor_rec2(sym, 0);
	
	if (!extra_vstruct) 
		file << indent << "return *this;\n";
	
	indent.indent_level--;
	file << indent << "}\n";
}

int CPPFile::output_struct_ctor_rec1(Struct *sym, int num)
{
	if (sym->get_super())
		num = output_struct_ctor_rec1(sym->get_super(), num);
	
	for (Struct::entries_iterator i = sym->entries_begin();
	     i != sym->entries_end(); ++i)
	{
		if (num++)
			file << ",\n" << indent;
		
		output_datum(sym, *i, -1);
		file << "_arg" << num;
	}
	
	return num;
}

int CPPFile::output_struct_ctor_rec2(Struct *sym, int num)
{
	if (sym->get_super())
		num = output_struct_ctor_rec2(sym->get_super(), num);
	
	for (Struct::entries_iterator i = sym->entries_begin();
	     i != sym->entries_end(); ++i)
	{
		Datum *d = *i;
		file << indent << **d->name << " = _arg" << ++num << ";\n";
	}
	
	return num;
}

void CPPFile::output(Struct *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			if (sym->is_virtual()) {
				output_nsdecl(sym, output_vstruct_ns);
			} else {
				output_nsdecl(sym);
			}

			break;
		
		case trav_forward: {
			bool nested = indent.indent_level != 0;

			if (!nested)
				ns_in(sym, "FWD_");
			else
				extra_newline();

			file << indent << "struct " << **sym->name << ";\n";
			
			if (!nested)
				ns_out(sym);
				
			break;
		}
		
		case trav_full: {
			output_pass(sym, trav_nsdecl);
			Struct *super = sym->get_super();

			if (super)
				output_pass(super, trav_full);
			
			if (sym->is_virtual())
				output_vstruct_info(sym);

			declare_dependencies(sym);
			ns_in(sym);

			file << indent << "struct " << get_definition_name(sym);

			if (super) {
				const String *supername = super->get_fq_name("_ns")
				                               ->flatten("::");
				file << " :\n" << indent << "public ::" << *supername;
			}
	
			file << "\n" << indent << "{\n";
			downscope();
			
			if (sym->is_virtual())
				output_vstruct_main(sym);

			int offset = 0;
			for (Struct::entries_iterator i = sym->entries_begin();
			     i != sym->entries_end(); ++i)
			{
				extra_newline();
				Datum *d = *i;
				file << indent;
				offset = output_datum(sym, d, offset);
				file << **d->name << ";\n";
			}
			
			bool empty_struct = true;
			for (Struct *s = sym; s; s = s->get_super()) {
				if (s->entries_begin() != s->entries_end()) {
					empty_struct = false;
					break;
				} 
			}
			
			if (!empty_struct) {
				output_struct_ctor(sym, false);
				
				if (sym->is_virtual())
					output_struct_ctor(sym, true);
			}

			upscope();
			file << indent << "};\n";
	
			ns_out(sym);
			
			output_aliases_and_types(sym);
			
			for (NameSpace::const_iterator i = sym->begin(); i != sym->end(); ++i) {
				Symbol *sym2 = (*i).second;
				output_pass(sym2, trav_full);
			}

			break;
		}
		
		default:
			BUG();
	}
}

void CPPFile::output_iface_ns(CPPFile *file, NameSpace *sym)
{
	Interface *i = dynamic_cast<Interface *>(sym);
	assert(i);
	
	file->output_guid(i->def.guid);
}

void CPPFile::output(Interface *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			output_nsdecl(sym, output_iface_ns);
			break;
		
		case trav_forward: {
			bool nested = indent.indent_level != 0;

			if (!nested)
				ns_in(sym, "FWD_");
			else
				extra_newline();

			file << indent << "struct " << **sym->name << ";\n"
			     << indent << "struct _i_" << **sym->name << ";\n";

			if (!nested)
				ns_out(sym);

			break;
		}
		
		case trav_obj_stub: {
			output_pass(sym, trav_forward);

			for (Interface::supers_iterator i = sym->supers_begin();
			     i != sym->supers_end(); ++i)
			{
				Interface *super = *i;
				output_pass(super, trav_obj_stub);
			}

			declare_dependencies(sym);
			ns_in(sym, "STUB_");
			
			file << indent << "struct "
			     << get_definition_name(sym) << " {\n";

			downscope();

			output_wrapper(sym);

			upscope();
			file << indent << "};\n";

			ns_out(sym);
			break;
		}
		
		case trav_obj_def: {
			output_pass(sym, trav_obj_stub);

			for (Interface::supers_iterator i = sym->supers_begin();
			     i != sym->supers_end(); ++i)
			{
				Interface *super = *i;
				output_pass(super, trav_full);
			}
			
			declare_dependencies(sym);
			ns_in(sym, "OBJDEF_");

			file << indent << "struct ";
			file << get_definition_name(sym, "_i_") << " {\n";

			downscope();

			output_internal(sym);

			upscope();
			file << indent << "};\n\n";

			output_casts(sym);

			ns_out(sym);
			break;
			
		case trav_full:
			output_pass(sym, trav_obj_def);

			declare_dependencies(sym, true);
			ns_in(sym);
			
			output_method_defs(sym);
			
			ns_out(sym);
			output_aliases_and_types(sym);

			for (NameSpace::const_iterator i = sym->begin(); i != sym->end(); ++i) {
				Symbol *sym2 = (*i).second;
				output_pass(sym2, trav_full);
			}
	
			break;
		
		default:
			BUG();
		}
	}
}

void CPPFile::output_bf_elem(Datum *d, int pos, int bits,
                             string &prefix)
{
	Type *t = d->type;
	BitField *bf = dynamic_cast<BitField *>(t);
	Enum *en = dynamic_cast<Enum *>(t);
	
	string fieldname;

	if (!d->name->compare(0, 4, "get_") ||
	    !d->name->compare(0, 4, "set_"))
		fieldname = '_';
	
	fieldname.append(**d->name);
	
	if (bf) {
		string newprefix(prefix);
		newprefix.append(**d->name);
		newprefix.append("_IDLNS_");
		output_bf(bf, d->def.icon, bits, newprefix);
		
		// FIXME: getters and setters
	} else if (en || !t) {
		file << indent << "uint" << bits << "_t "
		     << prefix << fieldname << ':' << d->def.icon << ";\n";
	} else {
		// This is checked here rather than in input.cc, because
		// I'm lazy.
		
		fprintf(stderr, "idlc: Bad input: \"%s\" cannot be the type of \"%s\"\n",
		        t->get_fq_name()->flatten()->c_str(),
		        d->get_fq_name()->flatten()->c_str());
	
		throw UserError();
	}
}

void CPPFile::output_bf(BitField *sym, int bits, int typebits,
                        string &prefix)
{
	int size = 0;
	
	assert(bits == sym->def.bits || bits == typebits);
	
	for (BitField::entries_iterator i = sym->entries_begin();
	     i != sym->entries_end(); ++i)
		size += (*i)->def.icon;
	
	if (size > sym->def.bits) {
		// FIXME: This isn't detected in the front end,
		// but even once it is, this should stay as a
		// replacement for input.cc checking.
		
		fprintf(stderr, "idlc: \"%s\" is too small (%d bits) for its "
		                "contents (%d bits)\n",
		        sym->get_fq_name()->flatten()->c_str(),
		        sym->def.bits, size);
	
		throw UserError();
	}
	
	if (target->bitfield_big_endian) {
		if (size != bits) {
			// The prefix is put at the end, so that we can avoid
			// consecutive underscores or an underscore followed
			// by a capital, both of which are reserved in C++.
		
			file << indent << "uint" << bits << "_t _pad_" << prefix
			     << ':' << bits - size << ";\n";
		}

		int pos = sym->def.bits;

		for (BitField::entries_reverse_iterator i = sym->entries_rbegin();
		     i != sym->entries_rend(); ++i)
		{
			Datum *d = *i;
			pos -= d->def.icon;
			output_bf_elem(d, pos, typebits, prefix);
		}
	} else {
		int pos = 0;
	
		for (BitField::entries_iterator i = sym->entries_begin();
		     i != sym->entries_end(); ++i)
		{
			Datum *d = *i;
			output_bf_elem(d, pos, typebits, prefix);
			pos += d->def.icon;
		}
	}
}

void CPPFile::output(BitField *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			output_nsdecl(sym);
			break;

		case trav_forward:
			extra_newline();
			file << indent << "union " << **sym->name << ";\n";
			break;
		
		case trav_full: {
			int bits = round_up_bits(sym->def.bits);
			ns_in(sym);

			file << indent << "union ";
			file << get_definition_name(sym) << " {\n";
			downscope();
			
			file << indent << "struct {\n";

			downscope();
			string nullprefix;
			output_bf(sym, bits, bits, nullprefix);
			upscope();
			
			file << indent << "};\n\n"
			     << indent << "uint" << bits << "_t _raw;\n\n"
			     << indent << **sym->name << "()\n"
			     << indent << "{\n"
			     << indent << "\t_raw = 0;\n"
			     << indent << "}\n\n"
			     << indent << **sym->name << "(uint" << bits << "_t _init)\n"
			     << indent << "{\n"
			     << indent << "\t_raw = _init;\n"
			     << indent << "}\n\n"
			     << indent << "operator uint" << bits << "_t()\n"
			     << indent << "{\n"
			     << indent << "\treturn _raw;\n"
			     << indent << "}\n";
			
			upscope();
			file << indent << "};\n";

			ns_out(sym);
			output_aliases_and_types(sym);
			break;
		}

		default:
			BUG();
	}
}


void CPPFile::output(Enum *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			/* no-op */
			break;

		case trav_forward: {
			bool do_ns_out = false;
			
			if (indent.indent_level == 0) {
				ns_in(sym);
				do_ns_out = true;
			} else {
				extra_newline();
			}
			
			file << indent << "struct " << **sym->name << " {\n";
			downscope();
			
			for (Enum::entries_iterator i = sym->entries_begin();
			     i != sym->entries_end(); ++i)
			{
				Datum *d = *i;
				
				file << indent << "static const uint" 
				     << round_up_bits(sym->def.bits) << "_t "
				     << **d->name << " = " << d->def.ucon << ";\n";
			}
			
			file << '\n' << indent << "unsigned int _val;\n\n";

			file << indent << **sym->name << "()\n"
			     << indent << "{\n"
			     << indent << "\t_val = 0;\n"
			     << indent << "}\n\n";

			file << indent << **sym->name << "(unsigned int val)\n"
			     << indent << "{\n"
			     << indent << "\t_val = val;\n"
			     << indent << "}\n\n";

			file << indent << "operator unsigned int()\n"
			     << indent << "{\n"
			     << indent << "\treturn _val;\n"
			     << indent << "}\n\n";
			
			upscope();
			file << indent << "};\n";

			if (do_ns_out)
				ns_out(sym);

			break;
		}
		
		case trav_full:
			// Nothing particular to do here, other than to make sure
			// that trav_forward has happened (which will always need to
			// be done here if it's not a nested type).
			
			output_pass(sym, trav_forward);
			break;

		default:
			BUG();
	}
}

void CPPFile::output(BasicType *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			/* no-op */
			break;

		case trav_forward: {
			bool do_ns_out = false;
			
			if (indent.indent_level == 0) {
				ns_in(sym);
				do_ns_out = true;
			} else {
				extra_newline();
			}
			
			file << indent << "typedef ";
			assert(!is_array(sym->def));
			cpp_output_type(file, sym->def, false);
			file << **sym->name << ";\n";

			if (do_ns_out)
				ns_out(sym);

			break;
		}
		
		case trav_full:
			output_pass(sym, trav_forward);
			break;

		default:
			BUG();
	}
}

void CPPFile::output(Alias *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
			/* no-op */
			break;

		case trav_forward: {
			bool do_ns_out = false;
			
			if (indent.indent_level == 0) {
				all_ns_in(sym);
				do_ns_out = true;
			} else {
				extra_newline();
			}
			
			const String *type = sym->get_concrete_sym()->get_fq_name("_ns")
			                        ->flatten("::");

			file << indent << "typedef " << *type << " " 
			     << **sym->name << ";\n";

			if (do_ns_out)
				all_ns_out(sym);

			break;
		}
		
		case trav_full:
			output_pass(sym, trav_forward);
			break;

		default:
			BUG();
	}
}

void CPPFile::output(TypeDef *sym, int pass, void *arg2)
{
	switch (pass) {
		case trav_nsdecl:
		case trav_forward:
			/* no-op */
			break;

		case trav_full: {
			output_pass(sym->get_concrete_sym(), trav_forward);
			
			bool do_ns_out = false;
			
			if (indent.indent_level == 0) {
				all_ns_in(sym);
				do_ns_out = true;
			} else {
				extra_newline();
			}
			
			const String *type = sym->get_concrete_sym()->get_fq_name("_ns")
			                        ->flatten("::");

			file << indent << "typedef " << *type << " " 
			     << **sym->name << ";\n";

			if (do_ns_out)
				all_ns_out(sym);

			break;
		}
		
		default:
			BUG();
	}
}

void CPPFile::output(Datum *sym, int pass, void *arg2)
{
	assert(sym->def.flags.field.Const);
	
	switch (pass) {
		case trav_nsdecl:
		case trav_forward:
			/* no-op */
			break;

		case trav_full: {
			if (sym->type)
				declare_type_dependency(sym->type, false);

			bool do_ns_out = false;
			
			if (indent.indent_level == 0) {
				all_ns_in(sym);
				do_ns_out = true;
			} else {
				extra_newline();
			}
			
			file << indent << "static const ";
			
			assert(!is_array(sym->def.basictype));
			
			if (sym->type)
				cpp_output_type(file, sym->type, false, false);
			else
				cpp_output_type(file, sym->def.basictype, false);

			file << **sym->name << " = ";
			
			CompiledBasicType *def;
			if (sym->type) {
				Symbol *real_type = sym->type->get_concrete_sym();
				BasicType *bt = dynamic_cast<BasicType *>(real_type);
				assert(bt);
				
				def = &bt->def;
			} else {
				def = &sym->def.basictype;
			}
			
			if (def->flags.field.Float) {
				file << sym->def.fcon;
			} else if (def->flags.field.Bool) {
				if (sym->def.ucon == 0)
					file << "false";
				else
					file << "true";
			} else {
				if (def->flags.field.Unsigned) {
					file << "0x" << std::hex << sym->def.ucon << std::dec << 'U';
				} else {
					file << sym->def.icon;
				}

				file << "LL";
			}
			
			file << ";\n";

			if (do_ns_out)
				all_ns_out(sym);

			break;
		}

		default:
			BUG();
	}
}

void CPPBinding::output_root(UserNameSpace *ns, const char *dir)
{
	delete new CPPFile(ns, dir);
}
