// util/heap.h -- A heap-based priority queue implementation
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#ifndef _UTIL_HEAP_H
#define _UTIL_HEAP_H

#include <assert.h>
#include <stddef.h>
#include <stdint.h>
#include <lowlevel/bitops.h>

namespace Util {
	// T must support operator < for key ordering, and
	// have a Heap<T>::Node member called heap_node.

	template <typename T>
	class Heap {
	public:
		struct Node {
			Node *parent, *left, *right;
			
			// A pointer to the parent's left or right pointer
			// (whichever side this node is on).  This is also
			// used for sanity checks (it is NULL for nodes not
			// on a queue).
			
			Node **parentlink;
			
			Node()
			{
				parent = left = right = NULL;
				parentlink = NULL;
			}
			
			bool is_on_heap()
			{
				return parentlink != NULL;
			}
		};
	
	private:
		Node *top;
		size_t num_nodes;
	
		T *node_to_type(Node *n)
		{
			return reinterpret_cast<T *>(reinterpret_cast<uintptr_t>(n) - 
			                             offsetof(T, heap_node));
		}
		
		// Swap n with its parent
		void swap(Node *n)
		{
			Node *p = n->parent;
			*p->parentlink = n;
			
			n->parent = p->parent;
			p->parent = n;
			
			Node **plink = p->parentlink;
			
			if (n->parentlink == &p->left) {
				assert(p->left == n);
			
				Node *tmp = p->right;
				p->right = n->right;
				n->right = tmp;
				
				p->left = n->left;
				n->left = p;
				
				p->parentlink = &n->left;
				
				if (n->right) {
					n->right->parent = n;
					n->right->parentlink = &n->right;
				}
			} else {
				assert(n->parentlink == &p->right);
				assert(p->right == n);

				Node *tmp = p->left;
				p->left = n->left;
				n->left = tmp;
				
				p->right = n->right;
				n->right = p;

				p->parentlink = &n->right;
				
				if (n->left) {
					n->left->parent = n;
					n->left->parentlink = &n->left;
				}
			}

			if (p->left) {
				p->left->parent = p;
				p->left->parentlink = &p->left;
	
				if (p->right) {
					p->right->parent = p;
					p->right->parentlink = &p->right;
				}
			} else {
				assert(!p->right);
			}
			
			n->parentlink = plink;
		}
		
		// Restore heap condition by moving n towards the root,
		// if required.

		void up(Node *n)
		{
			T *n_type = node_to_type(n);

			assert(!n->left || !(*node_to_type(n->left) < *n_type));
			assert(!n->right || !(*node_to_type(n->right) < *n_type));
			
			while (n->parent && *n_type < *node_to_type(n->parent)) {
				assert(!n->left || !(*node_to_type(n->left) < *n_type));
				assert(!n->right || !(*node_to_type(n->right) < *n_type));
		
				swap(n);
			}

			assert(!n->left || !(*node_to_type(n->left) < *n_type));
			assert(!n->right || !(*node_to_type(n->right) < *n_type));
		}
		
		// Restore heap condition by moving n towards the leaves,
		// if required.

		void down(Node *n)
		{
			T *n_type = node_to_type(n);

			assert(!n->parent || !(*n_type < *node_to_type(n->parent)));
			
			// If n has a right, it'll have a left.
			while (n->left) {
				assert(!n->parent || !(*n_type < *node_to_type(n->parent)));
				Node *child = n->left;
				
				if (n->right && 
				    *node_to_type(n->right) < *node_to_type(n->left))
					child = n->right;
			
				if (*node_to_type(child) < *n_type)
					swap(child);
				else
					break;
			}

			assert(!n->parent || !(*n_type < *node_to_type(n->parent)));
		}
		
		Node *get_by_index(size_t num, Node **parent, Node ***link)
		{
			assert(num_nodes == 0 || top);
			assert(num <= num_nodes);
			
			// It's easier to use 1-based indices, so that the high set bit
			// perfectly indicates which level the node is on.
			
			num += 1;
		
			Node *n = NULL;
			size_t bit = 1 << ll_get_order_round_down(num);
			*parent = NULL;
			Node **nextlink = &top;
			
			do {
				*link = nextlink;
				*parent = n;
				n = *nextlink;
				
				bit /= 2;
			
				if (num & bit)
					nextlink = &n->right;
				else
					nextlink = &n->left;
			} while (bit);
			
			return n;
		}

	public:
		Heap()
		{
			top = NULL;
			num_nodes = 0;
		}
		
		bool empty()
		{
			return num_nodes == 0;
		}
		
		T *get_top()
		{
			if (!top)
				return NULL;
			
			return node_to_type(top);
		}
	
		void add(T *t)
		{
			Node *n = &t->heap_node;
			assert(!n->is_on_heap());
			
			Node *ret = get_by_index(num_nodes, &n->parent, &n->parentlink);

			assert(ret == NULL);
			assert(num_nodes == 0 || n->parent);
			
			num_nodes++;
			*n->parentlink = n;
			n->left = n->right = NULL;
			
			up(n);
		}
		
		void del(T *t)
		{
			Node *n = &t->heap_node;
			assert(*n->parentlink == n);
			
			if (--num_nodes == 0) {
				top = NULL;
				n->parentlink = NULL;
				return;
			}
		
			Node *parent, **link;
			Node *last = get_by_index(num_nodes, &parent, &link);

			assert(last->parent);
			assert(last->is_on_heap());
			assert(!last->left);
			assert(!last->right);

			*link = NULL;
			*n->parentlink = last;
			*last = *n;
			n->parentlink = NULL;

			if (last->left) {
				last->left->parent = last;
				last->left->parentlink = &last->left;
			}
				
			if (last->right) {
				last->right->parent = last;
				last->right->parentlink = &last->right;
			}
			
			down(last);
		}
		
		void requeue(T *t) {
			Node *n = &t->heap_node;
			assert(*n->parentlink == n);
		
			if (!n->parent || *node_to_type(n->parent) < *t)
				down(n);
			else
				up(n);
		}
	};
}

#endif
