// arch/x64/descriptors.cc -- code to manage segment/trap/interrupt descriptors
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/types.h>
#include <arch/addrs.h>
#include <arch/thread.h>
#include <arch/mem.h>

namespace Arch {
namespace Priv {
	TSS tss;
}
}

using Arch::Priv::Descriptor;
using Arch::Priv::TSS;
using Arch::Priv::tss;

Descriptor x64_gdt[1024] = {
	{}, // The first entry is reserved for the NULL selector.
	{   // 0x08: data
		limit_low: 0xffff,
		base_low: 0,
		base_mid: 0,
		type: 2, // data segment, writable
		user: 1,
		dpl: 0,
		present: 1,
		limit_high: 0xff,
		sw: 0,
		code64: 0,
		opsize: 1,
		gran: 1,
		base_high: 0
	},
	{   // 0x10: 32-bit code
		limit_low: 0xffff,
		base_low: 0,
		base_mid: 0,
		type: 10, // code segment, readable
		user: 1,
		dpl: 0,
		present: 1,
		limit_high: 0xff,
		sw: 0,
		code64: 0,
		opsize: 1,
		gran: 1,
		base_high: 0
	},
	{   // 0x18: 64-bit code
		limit_low: 0xffff,
		base_low: 0,
		base_mid: 0,
		type: 10, // code segment, readable
		user: 1,
		dpl: 0,
		present: 1,
		limit_high: 0xff,
		sw: 0,
		code64: 1,
		opsize: 0,
		gran: 1,
		base_high: 0
	},
	{   // 0x20: TSS
		limit_low: sizeof(TSS) - 4,
		base_low: 0,
		base_mid: 0,
		type: 9,
		user: 0,
		dpl: 0,
		present: 1,
		limit_high: 0,
		sw: 0,
		code64: 0,
		opsize: 0,
		gran: 0,
		base_high: 0,
	},
	{   // 0x28: TSS high
	}
};

struct X64DescriptorTablePointer {
	u8 pad[6];
	u16 limit;
	u64 address;
};

X64DescriptorTablePointer x64_gdtr = {
	// G++ still won't handle complex labelled initializers, so
	// we have to explicitly initialize pad.
	
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(x64_gdt),
	address: reinterpret_cast<u64>(&x64_gdt)
};

X64DescriptorTablePointer x64_gdtr_phys = {
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(x64_gdt),
	address: reinterpret_cast<u64>(&x64_gdt) - KERNEL_START
};

struct X64InterruptDescriptor {
	u16 offset_low;
	u16 selector;
	u8 stack_index:3;
	u8 reserved:5;
	u8 type:4;
	u8 user:1;
	u8 dpl:2;
	u8 present:1;
	u16 offset_mid;
	u32 offset_high;
	u32 reserved2;
};
	
static X64InterruptDescriptor idt[256];

static X64DescriptorTablePointer idtr = {
	pad: { 0, 0, 0, 0, 0, 0 }, 
	limit: sizeof(idt),
	address: reinterpret_cast<u64>(&idt)
};

// Set a gate for INT num to start executing at addr.  
//
// If ints_off is set, then interrupts will remain disabled until
// software enables them; otherwise, the interrupt flag will have the
// same state as in the code that was interrupted.
//
// If user is true, then this gate can be called directly from
// userspace with the INT instruction.  Otherwise, usermode
// INT to this gate will cause a GPF, but it may still be reached
// from user code via hardware interrupt or exception.
// 
// If stack_index is non-zero, use the specified alternate stack even if
// interrupting kernel code.

static void set_int_gate(int num, void *addrptr, bool ints_off = false,
                         bool user = false, int stack_index = 0)
{
	u64 addr = (u64)addrptr;

	X64InterruptDescriptor desc = {
		offset_low: addr & 0xffff,
		selector: 0x18,
		stack_index: stack_index,
		reserved: 0,
		type: ints_off ? 14 : 15,
		user: 0,
		dpl: user ? 3 : 0,
		present: 1,
		offset_mid: (addr >> 16) & 0xffff,
		offset_high: addr >> 32
	};
	
	idt[num] = desc;
}

extern int x64_diverr, x64_gpf, x64_page_fault, x64_invalid_insn;
extern void *x64_irqs[256];

namespace Arch {
namespace Priv {
	void set_idt()
	{
		// GCC 4.0 pukes on "m" (&idtr.limit), saying it's not
		// directly addressable.
		
		asm volatile("lidtq 6(%0)" : : "r" (&idtr) : "memory");
		
		set_int_gate(0, &x64_diverr);
		set_int_gate(6, &x64_invalid_insn);
		set_int_gate(13, &x64_gpf);
		set_int_gate(14, &x64_page_fault, true);
		
		for (int i = 0x20; i < 0x30; i++)
			set_int_gate(i, x64_irqs[i]);
	}
}
}
