// arch/x64/mem.cc -- x64 misc. memory management
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/kernel.h>
#include <kern/mem.h>
#include <kern/pagealloc.h>
#include <kern/libc.h>

// Initial page tables have the first 4 MiB mapped, using large pages.

__attribute__((aligned(4096))) u64 x64_init_ptbl_l2[512] = {
	0x87,
	0x00200087,
};

// The ORing of 7 into these entries will be done in entry.S;
// doing it here causes the compiler to emit runtime code
// to do it.

__attribute__((aligned(4096))) u64 x64_init_ptbl_l3[512] = {
	reinterpret_cast<u64>(x64_init_ptbl_l2) - KERNEL_START
};

__attribute__((aligned(4096))) u64 x64_init_ptbl_l4[512] = {
	reinterpret_cast<u64>(x64_init_ptbl_l3) - KERNEL_START
};

extern int _end;

using Mem::PageAllocZone;

namespace Arch {
	size_t mem_end;
	
	PageAllocZone *pagezones[3];

	namespace Priv {
		PageAllocZone isadmazone, dma32zone, highzone;
	
		PageAllocZone *isadmazonelist[2] = { &isadmazone, NULL };
		PageAllocZone *dma32zonelist[3];
		PageAllocZone *normalzonelist[4];
	}
	
	PageAllocZone **pagezonelists[3] = { Priv::normalzonelist,
	                                     Priv::dma32zonelist,
	                                     Priv::isadmazonelist };

	uintptr_t next_free_bootmem = reinterpret_cast<uintptr_t>(&_end) -
	                              KERNEL_START + PHYSMEM_START;

	namespace Priv {
		void early_adjust_mappings()
		{
			using Mem::get_bootmem;
		
			// Clear low-address mappings and invalidate TLB
			x64_init_ptbl_l4[0] = 0;
			x64_init_ptbl_l3[0] = 0;
			asm volatile("movq %0, %%cr3" : : "r" (kvirt_to_phys(x64_init_ptbl_l4)));
			
			// Mark the ktext mapping global now that it's not mapped at address
			// zero.
			
			x64_init_ptbl_l2[0] |= 0x100;
			x64_init_ptbl_l2[1] |= 0x100;
			
			u64 l3phys = kvirt_to_phys(get_bootmem(page_size, page_size));
			u64 *l3 = static_cast<u64 *>(phys_to_ktext(l3phys));
			bzero(l3, page_size);
			x64_init_ptbl_l4[0x100] = l3phys | 7;
	
			u64 l2phys = kvirt_to_phys(get_bootmem(page_size, page_size));
			u64 *l2 = static_cast<u64 *>(phys_to_ktext(l2phys));
			bzero(l2, page_size);
			l3[0] = l2phys | 7;
	
			// Map at least as much as is mapped in ktext, so that
			// things like the VGA driver can use it early without
			// having to know about phys_to_ktext, and so map_physmem
			// doesn't need to special-case the use of phys_to_ktext
			// for the first couple pages.
			
			l2[0] = 0x187;
			l2[1] = 0x00200187;
		}
	
		void map_physmem()
		{
			using Mem::get_bootmem;
		
			// phys_to_ktext can be used for the first
			// 2MiB-minus-size-of-kernel of bootmem allocations.
		
			for (uintptr_t physpage = 2; physpage <= (mem_end - 1) / 512; physpage++)
			{
				uintptr_t virtpage = physpage + (PHYSMEM_START >> 21);
				
				u64 *l3;
				u64 l3phys = x64_init_ptbl_l4[(virtpage >> 18) & 511] &
				             ~(page_size - 1);
				
				if (!l3phys) {
					l3 = static_cast<u64 *>(get_bootmem(page_size, page_size));
					bzero(l3, page_size);
					x64_init_ptbl_l4[(virtpage >> 18) & 511] = 
						kvirt_to_phys(l3) | 7;
				} else {
					l3 = static_cast<u64 *>(phys_to_kvirt(l3phys));
				}
				
				u64 *l2;
				u64 l2phys = l3[(virtpage >> 9) & 511] & ~(page_size - 1);
	
				if (!l2phys) {
					l2 = static_cast<u64 *>(get_bootmem(page_size, page_size));
					bzero(l2, page_size);
					l3[(virtpage >> 9) & 511] = kvirt_to_phys(l2) | 7;
				} else {
					l2 = static_cast<u64 *>(phys_to_kvirt(l2phys));
				}
				
				l2[virtpage & 511] = (physpage << 21) | 0x187;
			}
			
			size_t pages_size = mem_end * sizeof(Mem::Page);
			Mem::pages = static_cast<Mem::Page *>(get_bootmem(pages_size, 8));
			Mem::last_page = Mem::pages + pages_size - 1;
			bzero(Mem::pages, pages_size);
			
			int normal = 0, dma = 0;
			uintptr_t highstart = highzonestart;
			uintptr_t dma32start = dma32zonestart;
			
			if (mem_end > highstart) {
				normalzonelist[normal++] = &highzone;
				highzone.init(highstart, mem_end - highstart);
			} else {
				highstart = mem_end;
			}
			
			if (mem_end > dma32start) {
				normalzonelist[normal++] = &dma32zone;
				dma32zonelist[dma++] = &dma32zone;
				dma32zone.init(dma32start, highstart - dma32start);
			} else {
				dma32start = mem_end;
			}
	
			normalzonelist[normal++] = &isadmazone;
			dma32zonelist[dma++] = &isadmazone;
			
			isadmazone.init(mem_start, dma32start);
		}
	}
}
