// arch/x64/multiboot.cc -- Code to interpret multiboot data and initialize
//                          free memory regions based thereupon.
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/libc.h>
#include <kern/pagealloc.h>

#include <arch/multiboot.h>
#include <arch/addrs.h>
#include <arch/paging.h>
#include <arch/mem.h>

#include <limits.h>

u32 x64_boot_info_phys;

namespace Arch {
namespace Priv {
namespace MultiBoot {
	BootInfo *boot_info;

	void parse_mmap()
	{
	}
	
	void make_mem_avail(uintptr_t page_start, uintptr_t page_end)
	{
		using Mem::pages;
			
		if (page_start < dma32zonestart) {
			size_t len;
			
			if (page_end <= dma32zonestart)
				len = page_end - page_start;
			else
				len = dma32zonestart - page_start;
			
			isadmazone.free(&pages[page_start], len);
			page_start = dma32zonestart;
		}

		if (page_start < highzonestart && page_end > dma32zonestart) {
			size_t len;
			
			if (page_end <= highzonestart)
				len = page_end - page_start;
			else
				len = highzonestart - page_start;
			
			dma32zone.free(&pages[page_start], len);
			page_start = highzonestart;
		}

		if (page_end > highzonestart)
			dma32zone.free(&pages[page_start], page_end - highzonestart);
	}

	void process_info()
	{
		if (x64_boot_info_phys > max_ktext_map) {
			printf("Cannot access boot info at %#.8x\n", x64_boot_info_phys);
			for(;;);
		}
		
		boot_info = (BootInfo *)phys_to_ktext(x64_boot_info_phys);
		mem_end = 0;

		if (!(boot_info->flags & (1 << BootInfo::flag_mmap))) {
			// FIXME: use mem_lower and mem_upper in this case.
			printf("MultiBoot info does not contain a memory map.\n");
			for(;;);
		}

		printf("BIOS Memory Map:\n");

		uint off = 0;
		while (off < boot_info->mmap_length) {
			u32 phys = boot_info->mmap_addr + off;
		
			if (phys > max_ktext_map) {
				printf("Cannot access BIOS memory map entry at %#.8x\n", phys);
				for(;;);
			}
			
			MemMap *mmap = (MemMap *)phys_to_ktext(phys);
			
			printf("0x%016llx - 0x%016llx, type %d\n",
			       mmap->base, mmap->base + mmap->len - 1, mmap->type);
		
			off += mmap->size + 4;
			
			if (mmap->type == MemMap::Available) {
				size_t page_end = (mmap->base + mmap->len) / page_size;
				
				if (page_end > mem_end)
					mem_end = page_end;
			}
		}
		
		map_physmem();
		
		off = 0;
		while (off < boot_info->mmap_length) {
			u32 phys = boot_info->mmap_addr + off;
			MemMap *mmap = (MemMap *)phys_to_ktext(phys);
			off += mmap->size + 4;
			
			if (mmap->type == MemMap::Available) {
				// Don't use any page that isn't fully in this entry.

				size_t page_start = (mmap->base + page_size - 1) / page_size;
				size_t page_end = (mmap->base + mmap->len) / page_size;
				

				// Don't make available any pages that overlap the
				// kernel text/data or bootmem allocations.
				
				uintptr_t kernelendpage = 
					(kvirt_to_phys((void *)next_free_bootmem) + page_size - 1) /
					page_size;
				
				// Don't overwrite page 1; the BIOS will need it for later calls in
				// emulation, and at least the BIOS I'm testing on doesn't mark
				// it as reserved.
		
				if (page_start == 0)
					page_start = 1;
				if (page_start >= page_end)
					continue;
				
				if (page_start <= kernelendpage && page_end >= 512) {
					if (page_start < 512)
						make_mem_avail(page_start, 512);
					
					if (page_end > kernelendpage)
						make_mem_avail(kernelendpage + 1, page_end);
				} else {
					make_mem_avail(page_start, page_end);
				}
			}
		}
	}
}
}
}
