// arch/x86/mem.cc -- x86 paging and misc. memory management
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/kernel.h>
#include <kern/mem.h>
#include <kern/pagealloc.h>
#include <kern/libc.h>

// Initial page tables have the first 4 MiB mapped, using large pages.

__attribute__((aligned(4096))) u32 x86_init_ptbl_l2[1024] = {
	0x87
};

extern int _end;

using Mem::PageAllocZone;

namespace Arch {
	u64 mem_end;
	
	namespace Priv {
		#define zonelist(x) ((x) * ((x) + 1) / 2 - 1)
		
		PageAllocZone pagezones[num_zones];
		PageAllocZone *pagezonelists_real[zonelist(num_zones + 1)];
	}
	
	PageAllocZone **pagezonelists[Priv::num_zones];

	uintptr_t next_free_bootmem = reinterpret_cast<uintptr_t>(&_end);
}

namespace Arch {
namespace Priv {
	void early_adjust_mappings()
	{
		using Mem::get_bootmem;
	
		// Clear low-address mapping and invalidate TLB
		x86_init_ptbl_l2[0] = 0;
		asm volatile("movl %0, %%cr3" : : "r" (kvirt_to_phys(x86_init_ptbl_l2)));
		
		// Mark the ktext mapping global now that it's not mapped at address
		// zero.  FIXME: check for and enable PGE
		
		x86_init_ptbl_l2[0x200] |= 0x100;
	}

	void map_physmem()
	{
		using Mem::get_bootmem;
	
		// phys_to_ktext can be used for the first
		// 4MiB-minus-size-of-kernel of bootmem allocations.
	
		for (uintptr_t physpage = 1; physpage <= (mem_end - 1) / (4096*1024);
		     physpage++)
		{
			uintptr_t virtpage = physpage + (PHYSMEM_START >> 22);
			x86_init_ptbl_l2[virtpage & 1023] = (physpage << 22) | 0x187;
		}
		
		size_t pages_size = (mem_end / page_size) * sizeof(Mem::Page);
		Mem::pages = static_cast<Mem::Page *>(get_bootmem(pages_size, 4));
		Mem::last_page = Mem::pages + pages_size - 1;
		bzero(Mem::pages, pages_size);
		
		int listpos[num_zones];
		
		for (int i = num_zones - 1; i >= 0; i--) {
			listpos[i] = zonelist(i);
			pagezonelists[num_zones - 1 - i] = &pagezonelists_real[listpos[i]];

			u64 rstart = mem_zone_regions[i].start;
			u64 rend = mem_zone_regions[i].end;
		
			if (mem_start <= rend && mem_end >= rstart) {
				if (rstart < mem_start)
					rstart = mem_start;
				if (rend > mem_end)
					rend = mem_end;
				
				ulong page_start = rstart / page_size;
				ulong page_len = (rend - rstart + 1) / page_size;
				
				pagezones[i].init(page_start, page_len);
				
				for (int j = i; j < num_zones; j++) {
					assert(listpos[j] < zonelist(j + 1));
					pagezonelists_real[listpos[j]++] = &pagezones[i];
				}
			}
		}
	}
}
}
