// arch/x86/multiboot.cc -- Code to interpret multiboot data and initialize
//                          free memory regions based thereupon.
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/libc.h>
#include <kern/pagealloc.h>

#include <arch/multiboot.h>
#include <arch/addrs.h>
#include <arch/paging.h>
#include <arch/mem.h>

#include <util/misc.h>
#include <limits.h>

u32 x86_boot_info_phys;
extern int _start;

namespace Arch {
namespace Priv {
namespace MultiBoot {
	BootInfo *boot_info;

	void parse_mmap()
	{
	}
	
	using Util::round_up;
	using Util::round_down;
	
	void make_mem_avail(u64 start, u64 end)
	{
		using Mem::pages;
		
		if (start > end)
			return;
		
		for (int i = 0; i < num_zones; i++) {
			u64 rstart = mem_zone_regions[i].start;
			u64 rend = mem_zone_regions[i].end;
		
			if (start <= rend && end >= rstart) {
				if (rstart < start)
					rstart = start;
				if (rend > end)
					rend = end;
				
				ulong page_len = (rend - rstart + 1) / page_size;
				pagezones[i].free(Mem::phys_to_page(rstart), page_len);
			}
		}
	}

	void process_info()
	{
		if (x86_boot_info_phys > max_ktext_map) {
			printf("Cannot access boot info at %#.8x\n", x86_boot_info_phys);
			for(;;);
		}
		
		boot_info = (BootInfo *)phys_to_kvirt(x86_boot_info_phys);
		mem_end = 0;
		
		if (!(boot_info->flags & (1 << BootInfo::flag_mmap))) {
			// FIXME: use mem_lower and mem_upper in this case.
			printf("MultiBoot info does not contain a memory map.\n");
			for(;;);
		}
	
		printf("BIOS Memory Map:\n");

		uint off = 0;
		while (off < boot_info->mmap_length) {
			u32 phys = boot_info->mmap_addr + off;
		
			if (phys > max_ktext_map) {
				printf("Cannot access BIOS memory map entry at %#.8x\n", phys);
				for(;;);
			}
			
			MemMap *mmap = (MemMap *)phys_to_kvirt(phys);
			
			printf("0x%016llx - 0x%016llx, type %d\n",
			       mmap->base, mmap->base + mmap->len - 1, mmap->type);
		
			off += mmap->size + 4;
			
			if (mmap->type == MemMap::Available) {
				u64 end = mmap->base + mmap->len;
				
				if (end > mem_end)
					mem_end = end;
			}
		}
		
		map_physmem();

		// Don't make available any pages that overlap the
		// kernel text/data or bootmem allocations.
			
		u64 kernelstart = round_down(kvirt_to_phys(&_start), page_shift);
		u64 kernelend = round_up(kvirt_to_phys((void *)next_free_bootmem),
		                         page_shift) - 1;
				
		off = 0;
		while (off < boot_info->mmap_length) {
			u32 phys = boot_info->mmap_addr + off;
			MemMap *mmap = (MemMap *)phys_to_kvirt(phys);
			off += mmap->size + 4;
			
			if (mmap->type == MemMap::Available) {
				// Don't use any page that isn't fully in this entry.

				u64 start = round_up(mmap->base, page_shift);
				u64 end = round_down(mmap->base + mmap->len, page_shift) - 1;
				
				// Don't overwrite page 1; the BIOS will need it for later calls in
				// emulation, and at least the BIOS I'm testing on doesn't mark
				// it as reserved.
		
				if (start < page_size)
					start = page_size;
				
				if (start <= kernelend && end >= kernelstart) {
					if (start < kernelstart)
						make_mem_avail(start, kernelstart - 1);
					
					if (end > kernelend)
						make_mem_avail(kernelend + 1, end);
				} else {
					make_mem_avail(start, end);
				}
			}
		}
	}
}
}
}
