// core/thread.cc -- Scheduler and thread creation/destruction
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <util/lock.h>
#include <kern/thread.h>

namespace Lock {
	// OPT: add inline, lockless uncontended case

	// FIXME: Allow a high-priority blocker to steal a lock from
	// a low-priority holder if the holder has not yet run.  This
	// prevents rescheduling every iteration if both threads are
	// repeatedly acquiring and releasing the lock.
	
	void Lock::lock()
	{
		DroppableAutoSpinLockRecIRQ autolock(spinlock);
		
		if (!lockval) {
			lockval = reinterpret_cast<ulong>(curthread);
			return;
		}
		
		assert(lockval != reinterpret_cast<ulong>(curthread));
		
		Threads::ThreadBlocker blocker(curthread);
		waitqueue.block(&blocker);
		
		if (!lockval) {
			lockval = reinterpret_cast<ulong>(curthread);
			waitqueue.unblock(&blocker);
			return;
		}
		
		autolock.unlock();
		curthread->block(&blocker);
		
		// FIXME: interruptible locks
		assert(lockval == reinterpret_cast<ulong>(curthread));
	}
	
	void Lock::unlock()
	{
		AutoSpinLockRecIRQ autolock(spinlock);
		assert(lockval == reinterpret_cast<ulong>(curthread));
		
		if (waitqueue.empty()) {
			lockval = 0;
			return;
		}
		
		Threads::Blocker *b = waitqueue.unblock_one();
		lockval = reinterpret_cast<ulong>(b->thread);
		b->wake();
	}
}
