// core/time.cc -- Clocks and timers
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/arch.h>
#include <kern/time.h>
#include <lowlevel/clock.h>
#include <System/Exceptions.h>

namespace Time {
	using namespace Lock;

	void MonotonicClock::get_time(Time *time)
	{
		if (!time)
			throw_idl(InvalidArgument, 0, nullarray);
	
		AutoSpinLockRecIRQ autolock(lock);
		s32 nanos = get_nanos(ll_getclock() - last_tick_llclock);
		
		*time = last_tick_time;
		time->nanos += nanos;
		
		while (time->nanos >= 1000000000) {
			time->nanos -= 1000000000;
			time->seconds++;
		}
	}
	
	void MonotonicClock::get_resolution(Time *res)
	{
		if (!res) 
			throw_idl(InvalidArgument, 0, nullarray);
		
		res->seconds = 0;
		
		if (llclocks_per_second >= 1000000000)
			res->nanos = 1;
		else
			res->nanos = 1000000000 / llclocks_per_second;
	}
	
	void MonotonicClock::tick()
	{
		unsigned long irq = lock.lock_recirq();
		
		while (ll_getclock() - last_tick_llclock > llclocks_per_tick) {
			last_tick_llclock += llclocks_per_tick;
			last_tick_time += clock_tick_increment;
		}
		
		Time next_tick = last_tick_time + clock_tick_increment;
		lock.unlock_recirq(irq);

		tick_timer.arm(next_tick);
	}

	void MonotonicClock::new_timer(Timer *timer)
	{
		if (!timer)
			throw_idl(InvalidArgument, 0, nullarray);
	
		// FIXME
		*timer = NULL;
		return;
	}
	
	void MonotonicClock::timer_tick(KTimerEntry *entry)
	{
		MonotonicClock *clock = static_cast<MonotonicClock *>(entry->data);
		clock->tick();
	}
	
	void MonotonicClock::init_tick_timer()
	{
		tick_timer.mux = monotonic_timers;
		tick_timer.func = timer_tick;
		tick_timer.data = this;

		last_tick_llclock = ll_getclock();
		last_tick_time.seconds = 0;
		last_tick_time.nanos = 0;
		
		Time next_tick = clock_tick_increment;
		tick_timer.arm(next_tick);
		
		Arch::timer_init();
	}
	
	void HWTimer::get_expiry(Time *EXPIRY, uint8_t *ARMED)
	{
		if (EXPIRY)
			*EXPIRY = expiry;
		if (ARMED)
			*ARMED = armed;
	}
	
	MonotonicClock monotonic_clock;
	TimerMux *monotonic_timers;
	HWTimer *hw_timer;
	
	void init()
	{
		hw_timer->calibrate();
		monotonic_timers = new TimerMux(monotonic_clock, *hw_timer);
		monotonic_clock.init_tick_timer();
	}
}

#include <servers/core/time/footer.cc>
