// Generic 64-bit PTE, mainly useful for non-process aspaces on 32-bit hosts,
// so that files, disks, etc. can address >4GiB.  64-bit hosts will probably
// want to use Arch::PTE instead, especially if direct copying between page
// tables of similar PTEs is implemented.  32-bit embedded systems may also
// wish to use Arch::PTE (for both less code and smaller page tables) if all
// storage can be addressed with 32 bits.

#ifndef _KERN_GENERIC_PTE_H
#define _KERN_GENERIC_PTE_H

#include <lowlevel/misc.h>
#include <lowlevel/atomic.h>

namespace Mem {
	struct GenDirPTE {
		typedef u64 VirtAddr;
		typedef ulong PhysAddr;
	
		void *raw;
		
		GenDirPTE() : raw(NULL)
		{
		}
		
		GenDirPTE(void *RAW) : raw(RAW)
		{
		}

		static uint addr_to_offset(VirtAddr addr, int shift)
		{
			int pages_per_table = Arch::page_size / sizeof(void *);
			return (addr >> shift) & (pages_per_table - 1);
		}
	
		void *get_table()
		{
			return raw;
		}

		static GenDirPTE set_table(void *addr)
		{
			return GenDirPTE(addr);
		}

		bool valid_pte()
		{
			return raw;
		}
		
		void set_pte(GenDirPTE *table, uint offset)
		{
			table[offset] = raw;
		}
		
		enum {
			shift_per_level = Arch::page_shift - _LL_LONG_LOGBYTES,
		};
	};

	union GenPTE {
		typedef u64 VirtAddr, PhysAddr;
		typedef GenDirPTE DirPTE;

		struct {
#ifdef BITFIELD_LE
			PhysAddr Valid:1;
			PhysAddr Writeable:1;
			PhysAddr Readable:1;
			PhysAddr Executable:1;
			PhysAddr User:1;
			PhysAddr Accessed:1;
			PhysAddr Dirty:1;
			PhysAddr FaultOnWrite:1;
			PhysAddr Addr:56;
#elif defined(BITFIELD_BE)
			PhysAddr Addr:56;
			PhysAddr FaultOnWrite:1;
			PhysAddr Dirty:1;
			PhysAddr Accessed:1;
			PhysAddr User:1;
			PhysAddr Executable:1;
			PhysAddr Readable:1;
			PhysAddr Writeable:1;
			PhysAddr Valid:1;
#else
#error Unspecified/unrecognized bitfield endianness
#endif
		};

		PhysAddr raw;

		GenPTE(PhysAddr init) : raw(init)
		{
		}
		
		GenPTE() : raw(0)
		{
		}

		operator PhysAddr()
		{
			return raw;
		}

		static uint addr_to_offset(VirtAddr addr, int shift)
		{
			int pages_per_table = page_size / sizeof(PhysAddr);
			return (addr >> shift) & (pages_per_table - 1);
		}
	
		PhysAddr pte_to_addr()
		{
			return raw & ~((u64)page_size - 1);
		}
	
		static GenPTE addr_to_pte(PhysAddr phys)
		{
			return phys & ~((u64)page_size - 1);
		}
		
		static void flags_to_pte(Mem::PTEFlags flagsin,
		                         Mem::PTEFlags maskin,
		                         GenPTE &flagsout,
		                         GenPTE &maskout)
		{
			flagsout = (PhysAddr)flagsin;
			maskout = (PhysAddr)maskin;
		}
		
		GenPTE set_flags(GenPTE mask, GenPTE flags)
		{
			return (raw & ~mask) | flags;
		}
	
		Mem::PTEFlags pte_to_flags()
		{
			return raw & (page_size - 1);
		}
	
		void set_pte(GenPTE *table, uint offset)
		{
			table[offset] = raw;
		}
		
		GenPTE xchg_pte(GenPTE *table, uint offset)
		{
			GenPTE old = table[offset];
			table[offset] = raw;
			return old;
		}
		
		bool valid_pte()
		{
			return Valid;
		}
		
		bool dirty_pte()
		{
			// FIXME: decide how to handle stacked dirty pages
			return false;
		}
		
		enum {
			page_size = Arch::page_size,
			page_shift = Arch::page_shift,
	
			pages_per_table = page_size / sizeof(PhysAddr),
			shift_per_level = page_shift - 3,
			num_levels = 2 + (64 - page_shift - shift_per_level - 1) /
			             DirPTE::shift_per_level,
			
			// kmap is ignored for stacked aspaces
			kmap_start = 0, kmap_end = 0
		};
	};
}

#endif
