// io/console/misc.cc -- Generic code relating to kernel debug consoles
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/console.h>
#include <kern/libc.h>
#include <util/spinlock.h>

namespace IO {
namespace Console {
	void Console::write(char *data, size_t len)
	{
		u64 len64 = len;
		write(Array<octet>((octet *)data, len), &len64);
	}
	
	// All consoles will set this to themselves; the last console
	// to be initialized will be used.  At some point, the console
	// used will be configurable (including the ability to broadcast
	// to multiple consoles).
	
	Console *primary_console;
}
}

static const size_t printf_buffer_size = 4096;
static char printf_buffer[printf_buffer_size];
Lock::SpinLock printf_lock;

// Note: this will not retry with a larger buffer if overflow
// happens.  If you really need that, you'll need to call snprintf
// and primary_console->write yourself.

void vga_write(uint8_t *buf, u64 len);

size_t printf(const char *str, ...)
{
	Lock::AutoSpinLockRecIRQ autolock(printf_lock);

	va_list args;
	va_start(args, str);
	size_t ret = vsnprintf(printf_buffer, printf_buffer_size, str, args);
	va_end(args);
	
	if (ret > printf_buffer_size)
		ret = printf_buffer_size;
	
	IO::Console::primary_console->write(printf_buffer, ret);
	return ret;
}
