// io/console/vga.cc -- Kernel debugging console for standard VGA text mode
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <System/IO.h>
#include <kern/kernel.h>
#include <kern/console.h>
#include <kern/notifier.h>
#include <kern/io.h>
#include <kern/libc.h>
#include <kern/addrs.h>
#include <util/lock.h>

void vga_write(char *str);

namespace IO {
namespace Console {
	class VGA : public Console {
		Lock::SpinLock lock;
	
		u16 *mem_base;
		unsigned long io_base;
		
		int line_width;
		int lines;
		int pos;
		
		void putc(int ch, int attr)
		{
			if (ch == '\n') {
				pos += line_width;
				pos -= pos % line_width;
			} else if (ch == '\r') {
				pos -= pos % line_width;
			} else {
				// Use ll_swap_le16 rather than store_le16 in the hopes
				// that the compiler can simply change the shifts around
				// rather than assembling it one way and then ll_swapping.
				//
				// It's done as one atomic store rather than two single
				// byte writes (which would eliminate endianness concerns)
				// to eliminate the possibility of a race where the video
				// card reads a character before its attribute has been
				// written.
				
				mem_base[pos++] = ll_swap_le16(ch | (attr << 8));
			}
			
			if (pos >= lines * line_width) {
				pos -= line_width;
				
				// FIXME: until memcpy is more than a stupid byte-at-a-time
				// loop, this could race with the video card.
				
				memcpy(mem_base, mem_base + line_width,
				       line_width * (lines - 1) * sizeof(u16));
				
				for (int i = 0; i < line_width; i++)
					mem_base[pos + i] = ll_swap_le16(' ' | (7 << 8));
			}
		}
		
		void get_cursor()
		{
			ll_out_8(0x3d4, 14);
			pos = (u16)ll_in_8(0x3d5) << 8;
			ll_out_8(0x3d4, 15);
			pos |= ll_in_8(0x3d5);
		}

		void set_cursor()
		{
			ll_out_8(0x3d4, 14);
			ll_out_8(0x3d5, pos >> 8);
			ll_out_8(0x3d4, 15);
			ll_out_8(0x3d5, pos & 0xff);
		}
		
	public:
		#include <servers/io/console/vga/IO/Console/VGA.h>
		
		VGA()
		{
			init_iface();

			// FIXME: should be PCI mem space, not phys mem.
			mem_base = (u16 *)Arch::phys_to_kvirt(0xb8000);
			io_base = 0x3d0;
			
			// FIXME: detect
			line_width = 80;
			lines = 25;

			primary_console = this;
		}
		
		void write(Array<octet> buf, u64 *LEN)
		{
			if (!valid_size_t(*LEN))
				/* throw something */
				return;
			
			size_t len = *LEN;
			
			if (len > buf.count)
				/* throw something */
				return;
		
			Lock::AutoSpinLockRecIRQ autolock(lock);
			get_cursor();
			
			for (size_t i = 0; i < len; i++)
				putc(buf.ptr[i], 7);

			set_cursor();			
		}
		
		void write_async(Array<octet> buf, u64 len, Notifier notifier)
		{
			write(buf, &len);
			io_notify(notifier, len, io_result::Success);
		}
	};
	
	// Use a static constructor, so that it can be used before
	// memory management is up and running.
	
	VGA vga;
}
}

#include <servers/io/console/vga/footer.cc>
