// io/timer/i8254.cc -- Intel 8254 and compatible timer driver
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kern/time.h>
#include <kern/io.h>
#include <lowlevel/clock.h>

namespace IO {
namespace Timer {
	using Time::HWTimer;
	using Time::monotonic_clock;
	using Time::hw_timer;
	using Time::Time;

	class I8254 : public HWTimer {
		Lock::SpinLock lock;
		bool armed;
	
		u16 read_time()
		{
			ll_out_8(0x43, 0);
			u16 ret = ll_in_8(0x40);
			ret |= (u16)ll_in_8(0x40) << 8;
			return ret;
		}
		
		void wait_for_zero()
		{
			u16 val;
		
			do {
				val = read_time();
			} while (val < 0x8000);

			do {
				val = read_time();
			} while (val > 0x8000);
		}
		
		enum {
			clock_freq = 1193182
		};
		
	public:
		I8254()
		{
			hw_timer = this;
			armed = false;
		}
	
		void calibrate()
		{
			wait_for_zero();
			s64 start_tick = ll_getclock();
			wait_for_zero();
			wait_for_zero();
			s64 end_tick = ll_getclock();

			s64 ticks_per_second = (end_tick - start_tick) * 
			                       (u64)clock_freq / 0x20000;
			
			monotonic_clock.calibrate(ticks_per_second);
		}
	
		void arm(Time new_expiry)
		{
			// Interrupts should always be disabled when
			// this is called.
			
			Lock::AutoSpinLock autolock(lock);

			// The 8254 is often slow to program, so don't re-program if the
			// expiry is the same.  To accomodate this, the tick timer must
			// be set to go off often enough that the expiry is never more
			// than 0xffff 8254 ticks (about 1/18.2 sec) in the future.

			if (armed && expiry != new_expiry) {
				Time now;
				monotonic_clock.get_time(&now);
				Time rel_expiry = new_expiry - now;
				expiry = new_expiry;
				
				u32 ticks;
				
				assert(rel_expiry.seconds <= 0);
	
				if (rel_expiry.seconds < 0)
					ticks = 1;
				else {
					// Add one extra tick to make sure we round up rather than
					// down; otherwise, we'll just end up programming the timer
					// for one tick and trying again.
					
					ticks = rel_expiry.nanos * (u64)clock_freq /
					        1000000000ULL + 1;
					
					assert(ticks <= 0xffff);
	
					if (ticks == 0)
						ticks = 1;
				}
				
				if (!armed)
					ll_out_8(0x43, 0x30);
					
				ll_out_8(0x40, ticks & 0xff);
				ll_out_8(0x40, ticks >> 8);
				armed = true;
			}
		}
		
		void disarm()
		{
			armed = false;
		}
	};
	
	I8254 i8254;
}
}
