// lib/kernel/time/timer.cc -- Generic timer multiplexing
//
// This software is copyright (c) 2006 Scott Wood <scott@buserror.net>.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal with
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following condition:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE
// SOFTWARE.

#include <kernel/time.h>
#include <lowlevel/clock.h>

namespace Time {
	void TimerMux::arm(TimerEntry *entry, Time expiry)
	{
		AutoSpinLockRecIRQ autolock(lock);
		
		bool was_empty = heap.empty();
		Time oldexpiry;

		if (!was_empty) {
			oldexpiry = heap.get_top()->expiry;
		} else {
			// Shut up the compiler
			oldexpiry.seconds = 0;
			oldexpiry.nanos = 0;
		}

		entry->expiry = expiry;
		
		if (entry->is_armed())
			heap.requeue(entry);
		else
			heap.add(entry);

		Time newexpiry = heap.get_top()->expiry;
		if (was_empty || oldexpiry != newexpiry)
			parent.arm(newexpiry);
	}

	void TimerMux::disarm(TimerEntry *entry)
	{
		AutoSpinLockRecIRQ autolock(lock);
		
		if (entry->is_armed()) {
			bool was_top = entry == heap.get_top();

			heap.del(entry);

			if (was_top)
				parent.arm(heap.get_top()->expiry);
		}
	}
	
	void TimerMux::run()
	{
		DroppableAutoSpinLockRecIRQ autolock(lock);
		
		while (true) {
			TimerEntry *top = heap.get_top();
			Time now;
			clock.get_time(&now);

			if (now < top->expiry)
				break;

			heap.del(top);

			autolock.unlock();
			top->action();
			autolock.lock();
		}
		
		if (!heap.empty())
			parent.arm(heap.get_top()->expiry);
	}
	
	void TimerEntry::set_action(System::Events::Event *event)
	{
		// FIXME
	}

	void TimerEntry::action()
	{
		// FIXME
	}
}

#include _KERNEL_SERVER(time/timer/footer.cc)
